local DepthFirst = require( 'Scripts.Depth-First' )

function love.load()	
	love.graphics.setBackgroundColor( 255, 255, 255 )
	love.math.noise( os.time() ) 
	
	Timer = .001
	Time = 0
	ShouldContinue = true
	
	Quads, Maze, QuadWidth, QuadHeight = LoadImageInformation()
	Width, Height = 80,60
	
	ScreenWidth, ScreenHeight = love.graphics.getWidth(), love.graphics.getHeight()
	ScaleX, ScaleY = ScreenWidth / ( Width * QuadWidth ), ScreenHeight / ( Height * QuadHeight )
	
	-- Mazes
	Starts, DepthMaze, VisitedCells = DepthFirst.NewStep( Width, Height, 4,{ { 1, 1 }, { 80, 1 }, { 1, 60 }, { 80, 60 } }  )
	
	StartX, StartY = nil, nil
	EndX, EndY = nil, nil
	
	-- References
	DepthReference = DepthFirst.DecodeMazeData( DepthMaze )
	
	-- Canvases
	love.graphics.setDefaultFilter( 'nearest', 'nearest' )
	DepthCanvas = SetToCanvas( DepthReference, Quads, Maze, QuadWidth, QuadHeight )
end

function love.draw()
	love.graphics.push()
		love.graphics.scale( ScaleX, ScaleY )				
		
		if StartX then 
			love.graphics.setColor( 0, 255, 0, 255 )
			love.graphics.rectangle( 'fill', ( StartX - 1 ) * QuadWidth, ( StartY - 1 ) * QuadHeight, QuadWidth, QuadHeight )
		end
		love.graphics.setColor( 255, 0, 0, 255 )
		if EndX then
			love.graphics.rectangle( 'fill', ( EndX - 1 ) * QuadWidth, ( EndY - 1 ) * QuadHeight, QuadWidth, QuadHeight )
		end
		
		for Index = 1, #Starts do
			local Start = Starts[Index]
			if ShouldContinue and Start[#Start].x and Start[#Start].y then
				love.graphics.rectangle( 'fill', ( Start[#Start].x - 1 ) * QuadWidth, ( Start[#Start].y - 1 ) * QuadHeight, QuadWidth, QuadHeight )
			
				love.graphics.setColor( 0, 255, 0, 255 )
				love.graphics.rectangle( 'fill', ( Start[1].x - 1 ) * QuadWidth, ( Start[1].y - 1 ) * QuadHeight, QuadWidth, QuadHeight )
			end
			
			love.graphics.setColor( 255, 255, 255, 255 )
			
			love.graphics.draw( DepthCanvas, 0, 0 )
			
			love.graphics.setColor( 255, 0, 0, 255 )
		end
		-- love.graphics.print( math.floor( love.mouse.getX() / ( QuadWidth * ScaleX ) ) + 1 .. ', ' .. math.floor( love.mouse.getY() / ( QuadHeight * ScaleY ) ) + 1, 1, 1 )
	love.graphics.pop()
end

function love.update( dt ) 
	Time = Time + dt
	if ShouldContinue and Time > Timer then
		for _ = 1, 1 do -- Using this, you can set how many times you want to call per frame- to speed up the process.
			local Stops = 0
			for Index = 1, #Starts do
				local Start = Starts[Index]
				if not DepthReference[ Start[#Start].y ] or not DepthReference[ Start[#Start].y ][ Start[#Start].x ] or not Start[ #Start ].ReverseAmount then Stops = Stops + 1 
				else 
					local x, y, ReverseAmount
					x, y, ReverseAmount, DepthMaze = DepthFirst.Step( Start[#Start].x, Start[#Start].y, DepthMaze, Start.VisitedCells, Start[#Start].ReverseAmount )
					Start[#Start + 1] = { x = x, y = y, ReverseAmount = ReverseAmount }
					DepthReference = DepthFirst.DecodeMazeData( DepthMaze )
				end
			end
			if Stops >= #Starts then End() end
		end
		
		DepthCanvas = SetToCanvas( DepthReference, Quads, Maze, QuadWidth, QuadHeight )
		Time = 0
	end 
end

function love.keypressed( Key, IsRepeat )
	if Key == 'escape' then love.event.quit() 
	elseif Key == ' ' then TakeScreenShot() 
	else
		if DepthFirst.UnvisitedCellsRemain( DepthMaze ) then 
			ShouldContinue = not ShouldContinue 
			if not ShouldContinue then love.window.setTitle( 'Paused' ) 
			else love.window.setTitle( 'Maze Creation Algorithms' ) end
		else 
			EndX, EndY = DepthFirst.ChooseRandomEndCell( DepthReference )
			StartX, StartY = DepthFirst.ChooseRandomEndCell( DepthReference )
		end
	end
end

function love.keyreleased( Key, IsRepeat )
end

function End()
	-- TakeScreenShot() 
	-- PrintMazeData( DepthReference )
	-- local LastCell = VisitedCells[#VisitedCells]
	-- print( LastCell.x, LastCell.y )
	DepthCanvas = SetToCanvas( DepthReference, Quads, Maze, QuadWidth, QuadHeight )
	ShouldContinue = false 
	love.window.setTitle( 'Maze Creation Algorithms' )
	EndX, EndY = DepthFirst.ChooseRandomEndCell( DepthReference )
	StartX, StartY = DepthFirst.ChooseRandomEndCell( DepthReference )
end

function TakeScreenShot()
	local Image = love.graphics.newScreenshot()
	love.filesystem.write( 'Maze.png', Image )
	Image:encode( 'Maze.png' )
end

function MakePO2( Dimension )
	local a = 1
	while a < Dimension do
		a = a * 2
	end
	return a
end

function SetToCanvas( Reference, Quads, Maze, QuadWidth, QuadHeight )
	local Canvas = love.graphics.newCanvas( MakePO2( ScreenWidth ), MakePO2( ScreenHeight ) )
	
	love.graphics.setCanvas( Canvas )
		Canvas:clear()
		
		for y = 1, #Reference do
			for x = 1, #Reference[y] do				
				love.graphics.draw( Maze, Quads[ Reference[y][x] ], ( x - 1 ) * QuadWidth, ( y - 1 ) * QuadHeight )
			end
		end
	love.graphics.setCanvas()
	
	return Canvas
end

function PrintMazeData( Reference )
	for y = 1, #Reference do
		for x = 1, #Reference[y] do
			if #Reference[y][x] < 2 then print( x, y, Reference[y][x] ) end
		end
	end
end

function LoadImageInformation()
	local Quads = {}
	
	local Maze = love.graphics.newImage( 'Images/Maze.png' )
	local QuadWidth, QuadHeight = 5, 5
	local QuadInformation = {
		L = { 0, 0 }, 
		LR = { 5, 0 }, 
		LU = { 10, 0 }, 
		LD = { 15, 0 }, 
		LRU = { 20, 0 }, 
		LRD = { 25, 0 }, 
		LUD = { 0, 5 }, 
		LRUD = { 5, 5 }, 
		R = { 10, 5 }, 
		U = { 15, 5 }, 
		D = { 20, 5 },
		RD = { 25, 5 }, 
		UD = { 0, 10 }, 
		RU = { 5, 10 }, 
		RUD = { 10, 10 }, 
	}
	QuadInformation[''] = { 15, 10 }
	for Index, Quad in pairs( QuadInformation ) do
		Quads[Index] = love.graphics.newQuad( QuadInformation[Index][1], QuadInformation[Index][2], QuadWidth, QuadHeight, Maze:getWidth(), Maze:getHeight() )
	end
	
	return Quads, Maze, QuadWidth, QuadHeight
end